"""Objects used in the Apibara client."""

from dataclasses import dataclass
from datetime import datetime
from typing import List, Literal, Optional, Union

import apibara.application.indexer_service_pb2 as indexer_service_pb2
from apibara.starknet import get_selector_from_name


@dataclass
class EventFilter:
    signature: str
    address: Optional[bytes]

    @classmethod
    def from_event_name(
        cls, name: str, address: Optional[Union[str, bytes]]
    ) -> "EventFilter":
        """Create an EventFilter from the event name."""
        if isinstance(address, str):
            address = bytes.fromhex(address.replace("0x", ""))
        return cls(name, address)

    @staticmethod
    def from_proto(p: indexer_service_pb2.EventFilter):
        address = None
        if len(p.address) > 0:
            address = p.address
        return EventFilter(signature=p.signature, address=address)

    def to_proto(self) -> indexer_service_pb2.EventFilter:
        return indexer_service_pb2.EventFilter(
            signature=self.signature,
            address=self.address,
        )

    def __str__(self) -> str:
        address = ""
        if self.address is not None:
            address = f", address=0x{self.address.hex()}"
        return f"EventFilter(signature={self.signature}{address})"


@dataclass
class Network:
    name: str
    type: Literal["evm", "starknet"]

    @staticmethod
    def from_proto(p: indexer_service_pb2.Network):
        if p.HasField("starknet"):
            return Network(p.starknet.name, "starknet")
        if p.HasField("ethereum"):
            return Network(p.ethereum.name, "evm")
        raise ValueError("invalid network proto object")

    def __str__(self) -> str:
        return f"Network(name={self.name}, type={self.type})"


@dataclass
class Indexer:
    id: str
    index_from_block: int
    indexed_to_block: int
    filters: List[EventFilter]
    network: Network

    @staticmethod
    def from_proto(p: indexer_service_pb2.Indexer):
        filters = [EventFilter.from_proto(f) for f in p.filters]
        network = Network.from_proto(p.network)
        return Indexer(p.id, p.index_from_block, p.indexed_to_block, filters, network)

    def __str__(self) -> str:
        return f'Indexer(id={self.id}, network={self.network}, blocks=[{self.index_from_block}, {self.indexed_to_block}], filters=[{", ".join(str(f) for f in self.filters)}])'


@dataclass
class Version:
    major: int
    minor: int
    patch: int

    @staticmethod
    def from_proto(p: indexer_service_pb2.Version):
        return Version(major=p.major, minor=p.minor, patch=p.patch)


@dataclass
class IndexerConnected:
    indexer: Indexer
    version: Version

    @staticmethod
    def from_proto(p: indexer_service_pb2.IndexerConnected):
        indexer = Indexer.from_proto(p.indexer)
        version = Version.from_proto(p.version)
        return IndexerConnected(indexer=indexer, version=version)


@dataclass
class BlockHeader:
    hash: bytes
    parent_hash: Optional[bytes]
    number: int
    timestamp: datetime

    @staticmethod
    def from_proto(p: indexer_service_pb2.BlockHeader):
        dt = datetime.fromtimestamp(p.timestamp.seconds)
        return BlockHeader(bytes(p.hash), bytes(p.parent_hash), p.number, dt)

    def __str__(self) -> str:
        return f"BlockHeader(hash=0x{self.hash.hex()}, parent_hash=0x{self.parent_hash.hex()}, number={self.number}, timestamp={self.timestamp})"


@dataclass
class Event:
    @staticmethod
    def from_proto(p: indexer_service_pb2.Event):
        if p.HasField("starknet"):
            return StarkNetEvent.from_proto(p.starknet)
        if p.HasField("ethereum"):
            return EthereumEvent.from_proto(p.ethereum)


@dataclass
class StarkNetEvent(Event):
    name: Optional[str]
    address: bytes
    log_index: int
    topics: List[bytes]
    data: List[bytes]
    transaction_hash: bytes

    @staticmethod
    def from_proto(p: indexer_service_pb2.StarkNetEvent):
        topics = [bytes(t.value) for t in p.topics]
        data = [bytes(d.value) for d in p.data]
        return StarkNetEvent(
            name=None,
            address=bytes(p.address),
            log_index=p.log_index,
            topics=topics,
            data=data,
            transaction_hash=bytes(p.transaction_hash),
        )

    def __str__(self) -> str:
        name = ""
        if self.name is not None:
            name = f"name={self.name}, "
        return f"StarkNetEvent({name}address=0x{self.address.hex()}, log_index={self.log_index}, ...{len(self.topics)} topics, ...{len(self.data)} data)"


@dataclass
class EthereumEvent(Event):
    name: Optional[str]
    address: bytes
    log_index: int
    topics: List[bytes]
    data: bytes
    transaction_hash: bytes

    @staticmethod
    def from_proto(p: indexer_service_pb2.EthereumEvent):
        topics = [bytes(t.value) for t in p.topics]
        return EthereumEvent(
            name=None,
            address=bytes(p.address),
            log_index=p.log_index,
            topics=topics,
            data=p.data,
            transaction_hash=bytes(p.transaction_hash),
        )

    def __str__(self) -> str:
        name = ""
        if self.name is not None:
            name = f"name={self.name}, "
        return f"EthereumEvent({name}address=0x{self.address.hex()}, log_index={self.log_index}, ...{len(self.topics)} topics, ...{len(self.data)} data)"


@dataclass
class NewBlock:
    new_head: BlockHeader

    @staticmethod
    def from_proto(p: indexer_service_pb2.NewBlock):
        new_head = BlockHeader.from_proto(p.new_head)
        return NewBlock(new_head)

    def __str__(self) -> str:
        return f"NewBlock(new_head={self.new_head})"


@dataclass
class Reorg:
    new_head: BlockHeader

    @staticmethod
    def from_proto(p: indexer_service_pb2.Reorg):
        new_head = BlockHeader.from_proto(p.new_head)
        return Reorg(new_head)

    def __str__(self) -> str:
        return f"Reorg(new_head={self.new_head})"


@dataclass
class NewEvents:
    block: BlockHeader
    events: List[Event]

    @staticmethod
    def from_proto(p: indexer_service_pb2.NewEvents):
        block = BlockHeader.from_proto(p.block)
        events = [Event.from_proto(ev) for ev in p.events]
        return NewEvents(block, events)

    def __str__(self) -> str:
        return f"NewEvents(block={self.block}, ...{len(self.events)} events)"



"""Apibara CLI. Manage applications and indexers from the command line."""

import asyncio
from email.policy import default
from functools import wraps

import click
import grpc
from click_help_colors import HelpColorsGroup

from apibara.client import Client
from apibara.model import EventFilter, Indexer


def async_command(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


@click.group(
    cls=HelpColorsGroup, help_headers_color="cyan", help_options_color="magenta"
)
def cli():
    pass


@cli.group()
def indexer():
    """Manage indexers."""
    pass


@indexer.command()
@click.argument("indexer-id", type=str)
@click.argument("event-name", type=str)
@click.option("--index-from-block", type=int, help="Start indexing from this block.")
@click.option("--address", type=str, help="Only index events emitted by this contract.")
@click.option("--server-url", type=str, default=None, help="Apibara server url.")
@async_command
async def create(
    indexer_id, event_name, index_from_block=None, address=None, server_url=None
):
    """Create a new indexer.

    The indexer is identified by its INDEXER_ID. By default, the indexer indexes all
    events with name EVENT_NAME from block 0.

    The indexer is not started after creation, you need to connect to it to start indexing.
    """
    async with Client.connect(server_url) as client:
        try:
            filter = EventFilter.from_event_name(event_name, address)
            new_indexer = await client.indexer_client().create_indexer(
                indexer_id, index_from_block, filter
            )
            _format_indexer(new_indexer)
        except Exception as ex:
            _format_exception(ex)


@indexer.command()
@click.option("--server-url", type=str, default=None, help="Apibara server url.")
@async_command
async def list(server_url=None):
    """List all available indexers."""
    async with Client.connect(server_url) as client:
        try:
            indexers = await client.indexer_client().list_indexer()
            for indexer in indexers:
                _format_indexer(indexer)
        except Exception as ex:
            _format_exception(ex)


@indexer.command()
@click.argument("indexer-id", type=str)
@click.option("--server-url", type=str, default=None, help="Apibara server url.")
@async_command
async def delete(indexer_id, server_url=None):
    """Delete the given indexer."""
    async with Client.connect(server_url) as client:
        try:
            indexer = await client.indexer_client().delete_indexer(indexer_id)
            _format_indexer(indexer)
        except Exception as ex:
            _format_exception(ex)


def _format_indexer(indexer: Indexer):
    click.secho(indexer.id, fg="cyan")
    blocks = click.style(
        f"[{indexer.index_from_block}, {indexer.indexed_to_block}]", fg="magenta"
    )
    click.echo(f"    blocks: {blocks}")
    click.echo(f"    filters:")
    for filter in indexer.filters:
        if filter.address == b"" or filter.address is None:
            address = "any"
        else:
            address = "0x" + filter.address.hex()
        address = click.style(address, fg="magenta")
        click.echo(f"    - address: {address}")
        click.echo(f"      topics:")
        for topic in filter.topics:
            topic = click.style(topic, fg="magenta")
            click.echo(f"      - {topic}")


def _format_exception(ex: Exception):
    if isinstance(ex, grpc.aio.AioRpcError):
        message = f"({ex.code()}) {ex.details()}"
    else:
        message = str(ex)
    click.secho(message, fg="red")



from typing import Any, List


class RpcClient:
    async def get_block_by_hash(self, hash: bytes) -> dict:
        pass

    async def get_block_by_number(self, number: int) -> dict:
        pass

    async def call(self, address: bytes, method: str, params: List[Any]) -> dict:
        pass



from eth_hash.auto import keccak

_MASK_250 = 2**250 - 1


def get_selector_from_name(func_name: str) -> int:
    """Get the selector for the given function name."""
    # This function comes from cairo-lang, but we want to
    # avoid that dependency to simplify installation.
    return _starknet_keccak(data=func_name.encode("ascii"))


def _starknet_keccak(data: bytes) -> int:
    return int.from_bytes(keccak(data), "big") & _MASK_250



"""Manage Apibara indexers."""

from contextlib import asynccontextmanager, contextmanager
from typing import AsyncIterator, List, Optional

from aiochannel import Channel
from grpc import StatusCode
from grpc.aio import AioRpcError

import apibara.application.indexer_service_pb2 as indexer_service_pb2
import apibara.application.indexer_service_pb2_grpc as indexer_service_pb2_grpc
from apibara.model import (
    EventFilter,
    Indexer,
    IndexerConnected,
    NewBlock,
    NewEvents,
    Reorg,
)


class IndexerClient:
    def __init__(self, channel: Channel) -> None:
        self._channel = channel
        self._stub = indexer_service_pb2_grpc.IndexerManagerStub(self._channel)

    async def get_indexer(self, id) -> Optional[Indexer]:
        try:
            response = await self._stub.GetIndexer(
                indexer_service_pb2.GetIndexerRequest(id=id)
            )
            if response and response.indexer and response.indexer.id:
                return Indexer.from_proto(response.indexer)
        except AioRpcError as ex:
            if ex.code() == StatusCode.NOT_FOUND:
                return None
            raise

    async def create_indexer(
        self,
        id: str,
        network_name: str,
        index_from_block: int,
        filters: List[EventFilter],
    ) -> Optional[Indexer]:
        filters = [f.to_proto() for f in filters]
        response = await self._stub.CreateIndexer(
            indexer_service_pb2.CreateIndexerRequest(
                id=id,
                network_name=network_name,
                index_from_block=index_from_block,
                filters=filters,
            )
        )

        if response and response.indexer:
            return Indexer.from_proto(response.indexer)

    async def delete_indexer(self, id: str) -> Optional[Indexer]:
        response = await self._stub.DeleteIndexer(
            indexer_service_pb2.DeleteIndexerRequest(id=id)
        )

        if response and response.indexer:
            return Indexer.from_proto(response.indexer)

    async def list_indexer(self) -> List[Indexer]:
        response = await self._stub.ListIndexer(
            indexer_service_pb2.ListIndexerRequest()
        )

        return [Indexer.from_proto(ix) for ix in response.indexers]

    @asynccontextmanager
    async def connect(self, indexer: Indexer) -> AsyncIterator["IndexerStream"]:
        # Create indexer stream. Channel is used to send messages
        # to the server.
        channel = Channel()
        response_iter = self._stub.ConnectIndexer(channel)
        stream = IndexerStream(channel, response_iter)

        # Connect and await for connected message.
        await stream._connect_indexer(indexer.id)
        connected_message = await stream.__anext__()

        if not isinstance(connected_message, IndexerConnected):
            raise RuntimeError(f"Could not connect indexer {indexer.id}")

        # Check server version with the one supported.
        # While in alpha it's acceptable to have breaking changes
        # with no backward compatibility.
        server_version = connected_message.version
        server_version = (server_version.major, server_version.minor)
        if server_version != (0, 3):
            raise RuntimeError(
                f"connected to apibara version {server_version}: not supported"
            )

        # Yield stream to the caller.
        yield stream

        # Cleanup channel after the user finished.
        channel.close()


class IndexerStream:
    def __init__(self, channel: Channel, iter) -> None:
        self._chan = channel
        self._iter = iter.__aiter__()

    async def ack_block(self, block_hash: bytes):
        ack = indexer_service_pb2.AckBlock(hash=block_hash)
        request = indexer_service_pb2.ConnectIndexerRequest(ack=ack)
        await self._chan.put(request)

    async def _connect_indexer(self, id: str):
        connect = indexer_service_pb2.ConnectIndexer(id=id)
        request = indexer_service_pb2.ConnectIndexerRequest(connect=connect)
        await self._chan.put(request)

    def __aiter__(self):
        return self

    async def __anext__(self):
        response = await self._iter.__anext__()
        return self._parse_response(response)

    def _parse_response(self, response):
        if response.HasField("connected"):
            return IndexerConnected.from_proto(response.connected)

        if response.HasField("new_block"):
            return NewBlock.from_proto(response.new_block)

        if response.HasField("reorg"):
            return Reorg.from_proto(response.reorg)

        if response.HasField("new_events"):
            return NewEvents.from_proto(response.new_events)

        raise RuntimeError(f"unknown ConnectIndexerResponse message:\n{response}")



"""Connect and interact with the Apibara server."""

from contextlib import asynccontextmanager
from typing import AsyncIterator, Optional

import grpc

from apibara.indexer.indexer import IndexerClient

_DEFAULT_APIBARA_SERVER_URL = "localhost:7171"


class Client:
    """Apibara client."""

    @staticmethod
    @asynccontextmanager
    async def connect(url: Optional[str] = None) -> AsyncIterator["Client"]:
        if url is None:
            url = _DEFAULT_APIBARA_SERVER_URL
        async with grpc.aio.insecure_channel(url) as channel:
            yield Client(channel)

    def __init__(self, channel) -> None:
        self._channel = channel

    def indexer_client(self) -> IndexerClient:
        return IndexerClient(self._channel)
